import logging
import os
from pyppeteer import connect, launch


log = logging.getLogger(__name__)


class Browser:
    """A wrapper around pyppeteer.browser.Browser."""

    endpoint = os.environ.get('PYPPETEER_BROWSER_ENDPOINT')

    def __init__(self, browser=None):
        self.browser = browser

    def __await__(self):
        return self.start().__await__()

    async def start(self):
        """Starts chrome process or connects to the existing chrome."""

        if self.browser:
            pass
        elif self.endpoint:
            browser_conn = {'browserWSEndpoint': self.endpoint}
            log.debug('connecting: {}'.format(browser_conn))
            self.browser = await connect(browser_conn)
        else:
            log.debug('launching new browser..')
            self.browser = await launch()

        return self

    async def page(self, url, force=False, cookies=()):
        """Makes new page and returns its object.

        Args:
            url (str): URL to navigate page to. The url should
                include scheme, e.g. `https://`.
            force (bool): `True` - to always return a new page.
            cookies (tuple): cookies for the page.

        Returns:
            page (pyppeteer.page.Page): page.

        """

        if not self.browser:
            await self.start()

        blank_page = None
        for page in await self.browser.pages():
            if page.url == 'about:blank':
                blank_page = page
            elif force:
                continue
            elif page.url == url:
                break
        else:
            page = blank_page or await self.browser.newPage()
            await page.setViewport({'width': 1200,  'height': 1920})
            await page.setCookie(*cookies)

            log.debug('go to %s ..' % url)
            await page.goto(url)

        return page



class Error(Exception):
    pass


class AuthError(Error):
    pass


class APIError(Error):
    def __init__(self, error):
        super().__init__(error)
        self.code = error['error_code']
        self.msg = error['error_msg']

    def __str__(self):
        return f'Error {self.code}: {self.msg}'


class APIScrapperError(Error):
    code = 0

    def __init__(self, msg):
        super().__init__(msg)
        self.msg = msg

    def __str__(self):
        return f'Error {self.code}: {self.msg}'


class CookieError(APIScrapperError):
    code = 1



from . import api, exceptions, objects, parser, sessions, utils

from .utils import parseaddr
from .exceptions import (
    Error,
    AuthError,
    APIError,
    APIScrapperError,
    CookieError,
)
from .sessions import (
    PublicSession,
    ClientSession,
    ServerSession,
    ImplicitClientSession,
    ImplicitServerSession,
)
from .api import API
import logging


logging.basicConfig()



import aiohttp
import asyncio
import hashlib
from yarl import URL

from .exceptions import Error, AuthError, APIError
from .parser import AuthPageParser
from .utils import full_scope, parseaddr, SignatureCircuit, Cookie


class Session:
    """A wrapper around aiohttp.ClientSession."""

    __slots__ = ('session', )

    def __init__(self, session=None):
        self.session = session or aiohttp.ClientSession()

    def __await__(self):
        return self.authorize().__await__()

    async def __aenter__(self):
        return await self.authorize()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def authorize(self):
        return self

    async def close(self):
        await self.session.close()


class PublicSession(Session):
    """Session for calling public API methods of Platform@Mail.Ru."""

    PUBLIC_URL = 'http://appsmail.ru/platform'
    CONTENT_TYPE = 'text/javascript; charset=utf-8'

    async def public_request(self, segments=(), params=None):
        """Requests public data.

        Args:
            segments (tuple): additional segments for URL path.
            params (dict): URL parameters.

        Returns:
            response (dict): JSON object response.

        """

        url = f'{self.PUBLIC_URL}/{"/".join(segments)}'

        try:
            async with self.session.get(url, params=params) as resp:
                status = resp.status
                response = await resp.json(content_type=self.CONTENT_TYPE)
        except aiohttp.ContentTypeError:
            raise Error(f'got non-REST path: {url}')

        if status != 200:
            raise APIError(response['error'])

        return response


class TokenSession(PublicSession):
    """Session for sending authorized requests."""

    API_URL = f'{PublicSession.PUBLIC_URL}/api'
    ERROR_MSG = 'See https://api.mail.ru/docs/guides/restapi/#sig.'

    __slots__ = ('app_id', 'private_key', 'secret_key', 'session_key', 'uid')

    def __init__(self, app_id, private_key, secret_key,
                 access_token, uid, cookies=(), session=None):
        super().__init__(session)
        self.app_id = app_id
        self.private_key = private_key
        self.secret_key = secret_key
        self.session_key = access_token
        self.uid = uid
        self.cookies = cookies

    @property
    def cookies(self):
        """HTTP cookies from cookie jar."""
        return [Cookie.from_morsel(m) for m in self.session.cookie_jar]

    @cookies.setter
    def cookies(self, cookies):
        loose_cookies = []

        for cookie in cookies:
            loose_cookie = Cookie.to_morsel(cookie)
            loose_cookies.append((loose_cookie.key, loose_cookie))

        self.session.cookie_jar.update_cookies(loose_cookies)

    @property
    def sig_circuit(self):
        if self.uid and self.private_key:
            return SignatureCircuit.CLIENT_SERVER
        elif self.secret_key:
            return SignatureCircuit.SERVER_SERVER
        else:
            return SignatureCircuit.UNDEFINED

    @property
    def required_params(self):
        """Required parameters."""
        params = {'app_id': self.app_id, 'session_key': self.session_key}
        if self.sig_circuit is SignatureCircuit.SERVER_SERVER:
            params['secure'] = '1'
        return params

    def params_to_str(self, params):
        query = ''.join(f'{k}={str(params[k])}' for k in sorted(params))

        if self.sig_circuit is SignatureCircuit.CLIENT_SERVER:
            return f'{self.uid}{query}{self.private_key}'
        elif self.sig_circuit is SignatureCircuit.SERVER_SERVER:
            return f'{query}{self.secret_key}'
        else:
            raise Error(self.ERROR_MSG)

    def sign_params(self, params):
        """Signs the request parameters according to signature circuit.

        Args:
            params (dict): request parameters

        Returns:
            sig (str): signature

        """

        query = self.params_to_str(params)
        sig = hashlib.md5(query.encode('UTF-8')).hexdigest()
        return sig

    async def request(self, segments=(), params=()):
        """Sends a request.

        Args:
            segments (tuple): additional segments for URL path.
            params (dict): URL parameters, contains key 'method', e.g.
                {
                    "method": "stream.getByAuthor",
                    "uid": "15410773191172635989",
                    "limit": 10,
                }

        Returns:
            response (dict): JSON object response.

        """

        url = f'{self.API_URL}/{"/".join(segments)}'

        params = {k: params[k] for k in params if params[k]}
        params.update(self.required_params)
        params.update({'sig': self.sign_params(params)})

        async with self.session.get(url, params=params) as resp:
            status = resp.status
            response = await resp.json(content_type=self.CONTENT_TYPE)

        if status != 200:
            raise APIError(response['error'])

        return response


class ClientSession(TokenSession):

    ERROR_MSG = 'Pass "uid" and "private_key" to use client-server circuit.'

    def __init__(self, app_id, private_key, access_token, uid,
                 cookies=(), session=None):
        super().__init__(app_id, private_key, '',
                         access_token, uid, cookies, session)


class ServerSession(TokenSession):

    ERROR_MSG = 'Pass "secret_key" to use server-server circuit.'

    def __init__(self, app_id, secret_key, access_token,
                 cookies=(), session=None):
        super().__init__(app_id, '', secret_key,
                         access_token, '', cookies, session)


class ImplicitSession(TokenSession):

    OAUTH_URL = 'https://connect.mail.ru/oauth/authorize'
    REDIRECT_URI = 'http%3A%2F%2Fconnect.mail.ru%2Foauth%2Fsuccess.html'

    __slots__ = ('email', 'passwd', 'scope',
                 'expires_in', 'refresh_token', 'token_type')

    def __init__(self, app_id, private_key, secret_key,
                 email, passwd, scope, session=None):
        super().__init__(app_id, private_key, secret_key, '', '', (), session)
        self.email = email
        self.passwd = passwd
        self.scope = scope or full_scope()

    @property
    def params(self):
        """Authorization parameters."""
        return {
            'client_id': self.app_id,
            'redirect_uri': self.REDIRECT_URI,
            'response_type': 'token',
            'scope': self.scope,
        }

    async def authorize(self, num_attempts=1, retry_interval=1):
        url, html = await self._get_auth_page()

        for attempt_num in range(num_attempts):
            if url.path.endswith('oauth/authorize'):
                url, html = await self._process_auth_form(html)

            if url.query.get('fail') == '1':
                raise AuthError('invalid login or password')

            if url.path.endswith('/oauth/success.html'):
                await self._get_auth_response()
                return self

            if attempt_num >= num_attempts:
                raise AuthError('Authorization failed')

            await asyncio.sleep(retry_interval)
            url, html = await self._get_auth_page()

    async def _get_auth_page(self):
        """Returns url and html code of authorization page."""

        async with self.session.get(self.OAUTH_URL, params=self.params) as resp:
            if resp.status != 200:
                raise AuthError("Wrong 'app_id' or 'scope'.")
            url, html = resp.url, await resp.text()

        return url, html

    async def _process_auth_form(self, html):
        """Submits a form with e-mail, password and domain
        to get access token and user id.

        Args:
            html (str): authorization page's html code.

        Returns:
            url (URL): redirected page's url.
            html (str): redirected page's html code.

        """

        parser = AuthPageParser()
        parser.feed(html)
        form_url, form_data = parser.form
        parser.close()

        domain, login = parseaddr(self.email)
        form_data['Login'] = login
        form_data['Domain'] = f'{domain}.ru'
        form_data['Password'] = self.passwd

        async with self.session.post(form_url, data=form_data) as resp:
            if resp.status != 200:
                raise AuthError('Failed to process authorization form')
            url, status, html = resp.url, resp.status, await resp.text()

        return url, html

    async def _get_auth_response(self):
        async with self.session.get(self.OAUTH_URL, params=self.params) as resp:
            location = URL(resp.history[-1].headers['Location'])
            url = URL(f'?{location.fragment}')

        try:
            self.session_key = url.query['access_token']
            self.expires_in = url.query['expires_in']
            self.refresh_token = url.query['refresh_token']
            self.token_type = url.query['token_type']
            self.uid = url.query['x_mailru_vid']
        except KeyError as e:
            key = e.args[0]
            raise AuthError(f'"{key}" is missing in the auth response')


class ImplicitClientSession(ImplicitSession):
    def __init__(self, app_id, private_key, email, passwd, scope, session=None):
        super().__init__(app_id, private_key, '', email, passwd, scope, session)


class ImplicitServerSession(ImplicitSession):
    def __init__(self, app_id, secret_key, email, passwd, scope, session=None):
        super().__init__(app_id, '', secret_key, email, passwd, scope, session)



from os.path import dirname, join
from setuptools import setup


readme_path = join(dirname(__file__), 'README.md')

with open(readme_path) as readme_file:
    readme = readme_file.read()


setup(
    name='aiomailru',
    version='0.0.4',
    author='Konstantin Togoi',
    author_email='konstantin.togoi@protonmail.com',
    url='https://github.com/KonstantinTogoi/aiomailru',
    description='Platform@Mail.ru Python REST API wrapper',
    long_description=readme,
    long_description_content_type='text/markdown',
    license='BSD',
    packages=['aiomailru', 'aiomailru.objects'],
    install_requires='aiohttp>=3.0.0',
    extra_require={
        'scrapers': ['pyppeteer<=0.0.25'],
    },
    keywords=['mail.ru api asyncio'],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from collections import defaultdict
import html.parser


class AuthPageParser(html.parser.HTMLParser):
    """Authorization page parser."""

    def __init__(self):
        super().__init__()
        self.inputs = {}
        self.url = ''

    @property
    def form(self):
        return self.url, self.inputs

    def handle_starttag(self, tag, attrs):
        attrs = defaultdict(str, attrs)

        if tag == 'input':
            if attrs['type'] != 'submit':
                self.inputs[attrs['name']] = attrs['value']
        elif tag == 'form':
            if attrs['method'] == 'post':
                self.url = attrs['action']



"""My.Mail.Ru scrapers."""

import asyncio
import logging
from concurrent.futures import FIRST_COMPLETED
from functools import lru_cache
from uuid import uuid4

from .exceptions import APIScrapperError, CookieError
from .api import API, APIMethod
from .browser import Browser
from .objects import Event, GroupItem
from .sessions import TokenSession


log = logging.getLogger(__name__)


class APIScraper(API, Browser):
    """API scraper."""

    __slots__ = ('browser', )

    def __init__(self, session: TokenSession, browser=None):
        API.__init__(self, session)
        Browser.__init__(self, browser)

    def __getattr__(self, name):
        return scrapers.get(name, APIMethod)(self, name)


class APIScraperMethod(APIMethod):
    """API scraper's method."""

    def __init__(self, api: APIScraper, name: str):
        super().__init__(api, name)

    def __getattr__(self, name):
        name = f'{self.name}.{name}'
        return scrapers.get(name, APIMethod)(self.api, name)

    async def __call__(self, scrape=False, **params):
        call = self.call if scrape else super().__call__
        return await call(**params)

    async def call(self, **params):
        raise NotImplementedError()


class GroupsGet(APIScraperMethod):
    """Returns a list of the communities to which the current user belongs."""

    url = 'https://my.mail.ru/my/communities'

    class Scripts:
        class Selectors:
            content = (
                'html body '
                'div.l-content '
                'div.l-content__center '
                'div.l-content__center__inner '
                'div.groups-catalog '
                'div.groups-catalog__mine-groups '
                'div.groups-catalog__small-groups '
            )
            bar = f'{content} div.groups-catalog__groups-more'
            catalog = f'{content} div.groups__container'
            button = f'{bar} span.ui-button-gray'
            item = f'{catalog} div.groups__item'

        click = f'document.querySelector("{Selectors.button}").click()'
        bar_css = 'n => n.getAttribute("style")'
        loaded = f'document.querySelectorAll("{Selectors.item}").length > {{}}'

    s = Scripts
    ss = Scripts.Selectors

    async def call(self, **params):
        cookies = self.api.session.cookies
        if not cookies:
            raise CookieError('Cookie jar is empty. Set cookies.')

        ext = params.get('ext', 0)
        limit = params.get('limit', 100)
        offset = params.get('offset', 0)
        page = await self.api.page(self.url, force=True, cookies=cookies)
        return await self.scrape(page, [], ext, limit, offset)

    async def scrape(self, page, groups, ext, limit, offset):
        """Appends groups from the `page` to the `groups` list."""

        _ = await page.screenshot()
        catalog = await page.J(self.ss.catalog)
        elements = await catalog.JJ(self.ss.item)

        start, stop = offset, min(offset + limit, len(elements))
        limit -= stop - start

        for i in range(start, stop):
            item = await GroupItem.from_element(elements[i])
            link = item['link'].lstrip('/')
            resp = await self.api.session.public_request([link])
            group, *_ = await self.api.users.getInfo(uids=resp['uid'])
            if ext:
                groups.append(group)
            else:
                groups.append(group['uid'])

        bar = await page.J(self.ss.bar)
        css = await page.Jeval(self.ss.bar, self.s.bar_css) or '' if bar else ''

        if limit == 0 or 'display: none;' in css:
            return groups
        else:
            await page.evaluate(self.s.click)
            await page.waitForFunction(self.s.loaded.format(len(groups)))
            return await self.scrape(page, groups, ext, limit, len(elements))


class GroupsGetInfo(APIScraperMethod):
    """Returns information about communities by their IDs."""

    class Scripts:
        class Selectors:
            main_page = (
                'html body '
                'div.l-content '
                'div.l-content__center '
                'div.l-content__center__inner '
                'div.b-community__main-page '
            )
            closed_signage = f'{main_page} div.mf_cc'

    s = Scripts
    ss = Scripts.Selectors

    async def call(self, **params):
        cookies = self.api.session.cookies
        if not cookies:
            raise CookieError('Cookie jar is empty. Set cookies.')

        uids = params['uids']
        info_list = await self.api.users.getInfo(uids=uids)

        for info in info_list:
            if 'group_info' in info:
                url = info['link']
                page = await self.api.page(url, force=True, cookies=cookies)
                group_info = await self.scrape(page)
                info['group_info'].update(group_info)

        return info_list

    async def scrape(self, page):
        """Returns additional information about a group.

        Object fields that are scraped here:
            - 'is_closed' - information whether the group's stream events
                are closed for current user.

        """

        signage = await page.J(self.ss.closed_signage)
        is_closed = True if signage is not None else False
        group_info = {'is_closed': is_closed}

        return group_info


class GroupsJoin(APIScraperMethod):
    """With this method you can join the group."""

    class Scripts:
        class Selectors:
            content = (
                'html body '
                'div.l-content '
                'div.l-content__center '
                'div.l-content__center__inner '
                'div.b-community__main-page '
                'div.profile div.profile__contentBlock '
            )
            links = (
                f'{content} '
                'div.profile__activeLinks '
                'div.profile__activeLinks_community '
            )
            join_span = f'{links} span.profile__activeLinks_button_enter'
            sent_span = f'{links} span.profile__activeLinks_link_modarated'
            approved_span = f'{links} span.profile__activeLinks_link_inGroup'
            auth_span = f'{links} div.l-popup_community-authorization'

        selector = 'document.querySelector("{}")'
        get_style = f'window.getComputedStyle({selector})'
        visible = f'{get_style}["display"] != "none"'
        join_span_visible = visible.format(Selectors.join_span)
        sent_span_visible = visible.format(Selectors.sent_span)
        approved_span_visible = visible.format(Selectors.approved_span)

        join_click = f'document.querySelector("{Selectors.join_span}").click();'

    s = Scripts
    ss = Scripts.Selectors

    async def call(self, **params):
        cookies = self.api.session.cookies
        if not cookies:
            raise CookieError('Cookie jar is empty. Set cookies.')

        group_id = params['group_id']
        info, *_ = await self.api.users.getInfo(uids=str(group_id))
        page = await self.api.page(info['link'], force=True, cookies=cookies)

        return await self.scrape(page)

    async def scrape(self, page):
        if await page.evaluate(self.s.join_span_visible):
            return await self.join(page)
        elif await page.evaluate(self.s.sent_span_visible):
            return 1
        elif await page.evaluate(self.s.approved_span_visible):
            return 1
        else:
            raise APIScrapperError('A join button not found.')

    async def join(self, page):
        await page.evaluate(self.s.join_click)
        await asyncio.sleep(1)

        if await page.evaluate(self.s.sent_span_visible):
            return 1
        elif await page.evaluate(self.s.approved_span_visible):
            return 1
        else:
            raise APIScrapperError('Failed to send join request.')


class StreamGetByAuthor(APIScraperMethod):
    """Returns a list of events from user or community stream by their IDs."""

    class Scripts:
        class Selectors:
            history = 'div[data-mru-fragment="home/history"]'
            event = 'div.b-history-event[data-astat]'

        class XPaths:
            history = (
                '/html/body/div[@id="boosterCanvas"]'
                '//div[@data-mru-fragment="home/history"]'
            )
            loaded = f'{history}[@data-state][@data-state!="loading"]'
            loading = f'{history}[@data-state="loading"]'

        scroll = 'window.scroll(0, document.body.scrollHeight)'
        state = 'n => n.getAttribute("data-state")'

    s = Scripts
    ss = Scripts.Selectors
    sx = Scripts.XPaths

    async def call(self, **params):
        uid = params.get('uid')
        skip = params.get('skip')
        limit = params.get('limit')
        uuid = skip if skip else uuid4().hex
        return await self.scrape(uid, skip, limit, uuid)

    @lru_cache(maxsize=None)
    async def scrape(self, uid, skip, limit, uuid):
        """Returns a list of events from user or community stream.

        Args:
            uid (int): User ID.
            skip (str): Latest event ID to skip.
            limit (int): Number of events to return.
            uuid (str): Unique identifier. May be used to prevent
                function from returning result from cache.

        Returns:
            events (list): Stream events.

        """

        cookies = self.api.session.cookies
        if not cookies:
            raise CookieError('Cookie jar is empty. Set cookies.')

        user, *_ = await self.api.users.getInfo(uids=uid)
        page = await self.api.page(user['link'], cookies=cookies)
        events = []

        async for event in self.stream(page):
            if skip:
                skip = skip if event['id'] != skip else False
            else:
                events.append(event)

            if len(events) >= limit:
                break

        return events

    async def stream(self, page):
        """Yields stream events from the beginning to the end.

        Args:
            page (pyppeteer.page.Page): Page with the stream.

        Yields:
            event (Event): Stream event.

        """

        history = await page.J(self.ss.history)
        history_ctx = history.executionContext
        state, elements = None, []

        while state != 'noevents':
            for element in await history.JJ(self.ss.event):
                yield await Event.from_element(element)

            await page.evaluate(self.s.scroll)

            tasks = [
                page.waitForXPath(self.sx.loading),
                page.waitForXPath(self.sx.loaded),
            ]

            _, pending = await asyncio.wait(tasks, return_when=FIRST_COMPLETED)

            for task in tasks:
                if not task.promise.done():
                    task.promise.set_result(None)
                    task._cleanup()

            for future in pending:
                future.cancel()

            state = await history_ctx.evaluate(self.s.state, history)
            if state == 'loading':
                await page.waitForXPath(self.sx.loaded)
                state = await history_ctx.evaluate(self.s.state, history)


scrapers = {
    'groups': APIScraperMethod,
    'groups.get': GroupsGet,
    'groups.getInfo': GroupsGetInfo,
    'groups.join': GroupsJoin,
    'stream': APIScraperMethod,
    'stream.getByAuthor': StreamGetByAuthor,
}
